@@ -1,12 +1,16 @@
 /*
  * conn.c
- * $Id: conn.c,v 1.25 2005/02/28 03:29:09 ianmacd Exp $
+ * $Id: conn.c,v 1.49 2006/04/20 16:39:46 ianmacd Exp $
  */
 
 #include "ruby.h"
 #include "rbldap.h"
-#include <sys/time.h>
-#include <unistd.h>
+#if defined(HAVE_SYS_TIME_H)
+# include <sys/time.h>
+#endif
+#if defined(HAVE_UNISTD_H)
+# include <unistd.h>
+#endif
 
 /* RDoc needs the following bogus code to find the parent module:
  *
@@ -15,6 +19,8 @@
 
 static VALUE rb_ldap_sort_obj = Qnil;
 extern VALUE rb_ldap_control_new2 (LDAPControl * ctl);
+extern VALUE rb_ldap_sslconn_initialize (int argc, VALUE argv[], VALUE self);
+extern VALUE rb_ldap_conn_rebind (VALUE self);
 
 VALUE rb_cLDAP_Conn;
 
@@ -55,7 +61,8 @@ rb_ldap_conn_s_allocate (VALUE klass)
 }
 
 /*
- * call-seq: LDAP::Conn.new(host='localhost', port=LDAP_PORT)  => LDAP::Conn
+ * call-seq:
+ * LDAP::Conn.new(host='localhost', port=LDAP_PORT)  => LDAP::Conn
  *
  * Return a new LDAP::Conn connection to the server, +host+, on port +port+.
  */
@@ -65,9 +72,10 @@ rb_ldap_conn_initialize (int argc, VALUE argv[], VALUE self)
   LDAP *cldap;
   char *chost;
   int cport;
+  int was_verbose = Qfalse;
   RB_LDAP_DATA *ldapdata;
 
-  VALUE arg1, arg2;
+  VALUE host, port;
 
   Data_Get_Struct (self, RB_LDAP_DATA, ldapdata);
   if (ldapdata->ldap)
@@ -75,7 +83,7 @@ rb_ldap_conn_initialize (int argc, VALUE argv[], VALUE self)
       return Qnil;
     }
 
-  switch (rb_scan_args (argc, argv, "02", &arg1, &arg2))
+  switch (rb_scan_args (argc, argv, "02", &host, &port))
     {
     case 0:
       chost = ALLOCA_N (char, strlen ("localhost") + 1);
@@ -83,12 +91,12 @@ rb_ldap_conn_initialize (int argc, VALUE argv[], VALUE self)
       cport = LDAP_PORT;
       break;
     case 1:
-      chost = StringValueCStr (arg1);
+      chost = StringValueCStr (host);
       cport = LDAP_PORT;
       break;
     case 2:
-      chost = StringValueCStr (arg1);
-      cport = NUM2INT (arg2);
+      chost = StringValueCStr (host);
+      cport = NUM2INT (port);
       break;
     default:
       rb_bug ("rb_ldap_conn_new");
@@ -99,13 +107,26 @@ rb_ldap_conn_initialize (int argc, VALUE argv[], VALUE self)
     rb_raise (rb_eLDAP_ResultError, "can't initialise an LDAP session");
   ldapdata->ldap = cldap;
 
-  rb_iv_set (self, "@sasl_quiet", Qfalse);
+  rb_iv_set (self, "@args", rb_ary_new4 (argc, argv));
+
+  /* Silence warning that next rb_iv_get produces. */
+  if (ruby_verbose == Qtrue)
+    {
+      was_verbose = Qtrue;
+      ruby_verbose = Qfalse;
+    }
+
+  if (rb_iv_get (self, "@sasl_quiet") != Qtrue)
+    rb_iv_set (self, "@sasl_quiet", Qfalse);
+  if (was_verbose == Qtrue)
+    ruby_verbose = Qtrue;
 
   return Qnil;
 };
 
 /*
- * call-seq: LDAP::Conn.open(host='localhost', port=LDAP_PORT)  => LDAP::Conn
+ * call-seq:
+ * LDAP::Conn.open(host='localhost', port=LDAP_PORT)  => LDAP::Conn
  *
  * Return a new LDAP::Conn connection to the server, +host+, on port +port+.
  */
@@ -116,10 +137,10 @@ rb_ldap_conn_s_open (int argc, VALUE argv[], VALUE klass)
   char *chost;
   int cport;
 
-  VALUE arg1, arg2;
+  VALUE host, port;
   VALUE conn;
 
-  switch (rb_scan_args (argc, argv, "02", &arg1, &arg2))
+  switch (rb_scan_args (argc, argv, "02", &host, &port))
     {
     case 0:
       chost = ALLOCA_N (char, strlen ("localhost") + 1);
@@ -127,12 +148,12 @@ rb_ldap_conn_s_open (int argc, VALUE argv[], VALUE klass)
       cport = LDAP_PORT;
       break;
     case 1:
-      chost = StringValueCStr (arg1);
+      chost = StringValueCStr (host);
       cport = LDAP_PORT;
       break;
     case 2:
-      chost = StringValueCStr (arg1);
-      cport = NUM2INT (arg2);
+      chost = StringValueCStr (host);
+      cport = NUM2INT (port);
       break;
     default:
       rb_bug ("rb_ldap_conn_new");
@@ -143,12 +164,12 @@ rb_ldap_conn_s_open (int argc, VALUE argv[], VALUE klass)
     rb_raise (rb_eLDAP_ResultError, "can't open an LDAP session");
   conn = rb_ldap_conn_new (klass, cldap);
 
-
   return conn;
 };
 
 /*
- * call-seq: conn.start_tls  => nil
+ * call-seq:
+ * conn.start_tls  => nil
  *
  * Initiate START_TLS for the connection, +conn+.
  */
@@ -160,7 +181,6 @@ rb_ldap_conn_start_tls_s (int argc, VALUE argv[], VALUE self)
   RB_LDAP_DATA *ldapdata;
   LDAPControl **serverctrls;
   LDAPControl **clientctrls;
-  int version = 3;
 
   switch (rb_scan_args (argc, argv, "02", &arg1, &arg2))
     {
@@ -184,6 +204,19 @@ rb_ldap_conn_start_tls_s (int argc, VALUE argv[], VALUE self)
   return Qnil;
 };
 
+VALUE
+rb_ldap_conn_rebind (VALUE self)
+{
+  VALUE ary = rb_iv_get (self, "@args");
+
+  if (rb_obj_is_kind_of (self, rb_cLDAP_SSLConn) == Qtrue)
+    return rb_ldap_sslconn_initialize (RARRAY (ary)->len, RARRAY (ary)->ptr,
+				       self);
+  else
+    return rb_ldap_conn_initialize (RARRAY (ary)->len, RARRAY (ary)->ptr,
+				    self);
+}
+
 /*
  * call-seq:
  * conn.simple_bind(dn=nil, password=nil)  => self
@@ -201,7 +234,21 @@ rb_ldap_conn_simple_bind_s (int argc, VALUE argv[], VALUE self)
   char *dn = NULL;
   char *passwd = NULL;
 
-  GET_LDAP_DATA (self, ldapdata);
+  Data_Get_Struct (self, RB_LDAP_DATA, ldapdata);
+  if (!ldapdata->ldap)
+    {
+      if (rb_iv_get (self, "@args") != Qnil)
+	{
+	  rb_ldap_conn_rebind (self);
+	  GET_LDAP_DATA (self, ldapdata);
+	}
+      else
+	{
+	  rb_raise (rb_eLDAP_InvalidDataError,
+		    "The LDAP handler has already unbound.");
+	}
+    }
+
   if (ldapdata->bind)
     {
       rb_raise (rb_eLDAP_Error, "already bound.");
@@ -278,7 +325,21 @@ rb_ldap_conn_bind_s (int argc, VALUE argv[], VALUE self)
   char *passwd = NULL;
   int method = LDAP_AUTH_SIMPLE;
 
-  GET_LDAP_DATA (self, ldapdata);
+  Data_Get_Struct (self, RB_LDAP_DATA, ldapdata);
+  if (!ldapdata->ldap)
+    {
+      if (rb_iv_get (self, "@args") != Qnil)
+	{
+	  rb_ldap_conn_rebind (self);
+	  GET_LDAP_DATA (self, ldapdata);
+	}
+      else
+	{
+	  rb_raise (rb_eLDAP_InvalidDataError,
+		    "The LDAP handler has already unbound.");
+	}
+    }
+
   if (ldapdata->bind)
     {
       rb_raise (rb_eLDAP_Error, "already bound.");
@@ -319,7 +380,8 @@ rb_ldap_conn_bind_s (int argc, VALUE argv[], VALUE self)
 };
 
 /*
- * call-seq: conn.unbind  => nil
+ * call-seq:
+ * conn.unbind  => nil
  *
  * Unbind the LDAP connection from the server.
  */
@@ -328,8 +390,6 @@ rb_ldap_conn_unbind (VALUE self)
 {
   RB_LDAP_DATA *ldapdata;
 
-  VALUE result;
-
   GET_LDAP_DATA (self, ldapdata);
   ldapdata->err = ldap_unbind (ldapdata->ldap);
   ldapdata->bind = 0;
@@ -340,7 +400,8 @@ rb_ldap_conn_unbind (VALUE self)
 };
 
 /*
- * call-seq: conn.bound?  => true or false
+ * call-seq:
+ * conn.bound?  => true or false
  *
  * Return *true* if the LDAP connection is still bound.
  */
@@ -349,13 +410,14 @@ rb_ldap_conn_bound (VALUE self)
 {
   RB_LDAP_DATA *ldapdata;
 
-  GET_LDAP_DATA (self, ldapdata);
+  Data_Get_Struct (self, RB_LDAP_DATA, ldapdata);
 
   return ldapdata->bind == 0 ? Qfalse : Qtrue;
 };
 
 /*
- * call-seq: conn.set_option(option, data)  => self
+ * call-seq:
+ * conn.set_option(option, data)  => self
  *
  * Set a session-wide option for this LDAP connection.
  *
@@ -476,7 +538,8 @@ rb_ldap_conn_s_set_option (VALUE klass, VALUE opt, VALUE data)
   return rb_ldap_conn_set_option (Qnil, opt, data);
 }
 
-/* call-seq: conn.get_option(opt)  => String
+/* call-seq:
+ * conn.get_option(opt)  => String
  *
  * Return the value associated with the option, +opt+.
  */
@@ -604,7 +667,8 @@ rb_ldap_conn_s_get_option (VALUE klass, VALUE opt)
 }
 
 /*
- * call-seq: conn.perror(msg)  => nil
+ * call-seq:
+ * conn.perror(msg)  => nil
  *
  * Print the text string associated with the error code of the last LDAP
  * operation. +msg+ is used to prefix the error.
@@ -614,7 +678,9 @@ rb_ldap_conn_perror (VALUE self, VALUE msg)
 {
   RB_LDAP_DATA *ldapdata;
   char *cmsg;
+#if (! defined(HAVE_LDAP_PERROR)) || defined(USE_NETSCAPE_SDK)
   char *str;
+#endif
 
   GET_LDAP_DATA (self, ldapdata);
   cmsg = StringValueCStr (msg);
@@ -629,7 +695,8 @@ rb_ldap_conn_perror (VALUE self, VALUE msg)
 };
 
 /*
- * call-seq: conn.result2error(msg)  => Fixnum
+ * call-seq:
+ * conn.result2error(msg)  => Fixnum
  *
  * Return the error code associated with the LDAP message, +msg+.
  */
@@ -649,7 +716,8 @@ rb_ldap_conn_result2error (VALUE self, VALUE msg)
 };
 
 /*
- * call-seq: conn.err2string(err)  => String
+ * call-seq:
+ * conn.err2string(err)  => String
  *
  * Return the text string associated with the LDAP error, +err+.
  */
@@ -669,7 +737,6 @@ VALUE
 rb_ldap_conn_get_errno (VALUE self)
 {
   RB_LDAP_DATA *ldapdata;
-  int cerr;
   VALUE err;
 
   GET_LDAP_DATA (self, ldapdata);
@@ -730,7 +797,6 @@ rb_ldap_conn_search_i (int argc, VALUE argv[], VALUE self,
 {
   VALUE base, scope, filter, attrs, attrsonly, sec, usec, s_attr, s_proc;
 
-  RB_LDAPENTRY_DATA *edata;
   LDAP *cldap;
   char *cbase;
   int cscope;
@@ -738,8 +804,7 @@ rb_ldap_conn_search_i (int argc, VALUE argv[], VALUE self,
   char **cattrs;
   char *sort_attr;
   int cattrsonly;
-  VALUE msg;
-  int i, j;
+  int i;
   struct timeval tv;
 
   GET_LDAP_DATA (self, (*ldapdata));
@@ -771,8 +836,17 @@ rb_ldap_conn_search_i (int argc, VALUE argv[], VALUE self,
     case 5:
       cattrsonly = (attrsonly == Qtrue) ? 1 : 0;
     case 4:
-      if (TYPE (attrs) == T_ARRAY)
+      if (TYPE (attrs) == T_NIL)
 	{
+	  cattrs = NULL;
+	}
+      else
+	{
+	  if (TYPE (attrs) == T_STRING)
+	    attrs = rb_ary_to_ary (attrs);
+	  else
+	    Check_Type (attrs, T_ARRAY);
+
 	  if (RARRAY (attrs)->len == 0)
 	    {
 	      cattrs = NULL;
@@ -787,10 +861,6 @@ rb_ldap_conn_search_i (int argc, VALUE argv[], VALUE self,
 	      cattrs[RARRAY (attrs)->len] = NULL;
 	    }
 	}
-      else
-	{
-	  cattrs = NULL;
-	};
     case 3:
       cbase = StringValueCStr (base);
       cscope = NUM2INT (scope);
@@ -813,7 +883,7 @@ rb_ldap_conn_search_i (int argc, VALUE argv[], VALUE self,
     }
   if (!(cmsg && (*cmsg)))
     {
-      rb_raise (rb_eRuntimeError, "no message");
+      rb_raise (rb_eRuntimeError, "no result returned by search");
     }
   Check_LDAP_Result ((*ldapdata)->err);
 
@@ -847,6 +917,31 @@ rb_ldap_conn_search_b (VALUE rdata)
   return Qnil;
 }
 
+static VALUE
+rb_ldap_conn_search2_b (VALUE rdata)
+{
+  void **data = (void *) rdata;
+  LDAP *cldap = (LDAP *) data[0];
+  LDAPMessage *cmsg = (LDAPMessage *) data[1];
+  VALUE ary = (VALUE) data[2];
+  LDAPMessage *e;
+  VALUE m;
+  VALUE hash;
+
+  for (e = ldap_first_entry (cldap, cmsg); e != NULL;
+       e = ldap_next_entry (cldap, e))
+    {
+      m = rb_ldap_entry_new (cldap, e);
+      hash = rb_ldap_entry_to_hash (m);
+      rb_ary_push (ary, hash);
+      if (rb_block_given_p ())
+	{
+	  rb_ensure (rb_yield, hash, rb_ldap_conn_invalidate_entry, m);
+	}
+    }
+  return Qnil;
+}
+
 static VALUE
 rb_ldap_msgfree (VALUE data)
 {
@@ -903,7 +998,9 @@ rb_ldap_parse_result (LDAP * cldap, LDAPMessage * cmsg)
  * search filter of +filter+.
  *
  * If +attrs+ is present, it should be an array of the attributes that the
- * search should return. By default, all attributes are returned.
+ * search should return. By default, all attributes are returned, which is the
+ * same as specifying an empty array or *nil*. Alternatively, +attrs+ may be a
+ * single string, in which case it will be treated as a single element array.
  *
  * If +attrsonly+ is *true*, attributes will be returned, but not their values.
  *
@@ -913,6 +1010,12 @@ rb_ldap_parse_result (LDAP * cldap, LDAPMessage * cmsg)
  * If +s_attr+ is given, it specifies the attribute on which to sort the
  * entries returned by the server. If +s_proc+ is given, it specifies a Proc
  * object that will be used to sort the entries returned by the server.
+ *
+ * Note that not all results may be returned by this method. If a
+ * size limit has been set for the number of results to be returned and this
+ * limit is exceeded, the results set will be truncated. You can check for
+ * this by calling LDAP::Conn#err immediately after this method and comparing
+ * the result to LDAP::LDAP_SIZELIMIT_EXCEEDED.
  */
 VALUE
 rb_ldap_conn_search_s (int argc, VALUE argv[], VALUE self)
@@ -920,13 +1023,13 @@ rb_ldap_conn_search_s (int argc, VALUE argv[], VALUE self)
   RB_LDAP_DATA *ldapdata;
   LDAPMessage *cmsg;
   LDAP *cldap;
-  LDAPMessage *e;
   VALUE rc_ary = Qnil;
 
   rb_ldap_conn_search_i (argc, argv, self, &ldapdata, &cmsg);
   cldap = ldapdata->ldap;
 
-  if (ldapdata->err == LDAP_SUCCESS)
+  if (ldapdata->err == LDAP_SUCCESS
+      || ldapdata->err == LDAP_SIZELIMIT_EXCEEDED)
     {
       void *pass_data[] = { (void *) cldap, (void *) cmsg };
 
@@ -941,41 +1044,20 @@ rb_ldap_conn_search_s (int argc, VALUE argv[], VALUE self)
   return self;
 }
 
-static VALUE
-rb_ldap_conn_search2_b (VALUE rdata)
-{
-  void **data = (void *) rdata;
-  LDAP *cldap = (LDAP *) data[0];
-  LDAPMessage *cmsg = (LDAPMessage *) data[1];
-  VALUE ary = (VALUE) data[2];
-  LDAPMessage *e;
-  VALUE m;
-  VALUE hash;
-
-  for (e = ldap_first_entry (cldap, cmsg); e != NULL;
-       e = ldap_next_entry (cldap, e))
-    {
-      m = rb_ldap_entry_new (cldap, e);
-      hash = rb_ldap_entry_to_hash (m);
-      rb_ary_push (ary, hash);
-      if (rb_block_given_p ())
-	{
-	  rb_ensure (rb_yield, hash, rb_ldap_conn_invalidate_entry, m);
-	}
-    }
-  return Qnil;
-}
-
 /*
  * call-seq:
  * conn.search2(base_dn, scope, filter, attrs=nil, attrsonly=false,
+ *		sec=0, usec=0, s_attr=nil, s_proc=nil)  => array
+ * conn.search2(base_dn, scope, filter, attrs=nil, attrsonly=false,
  *		sec=0, usec=0, s_attr=nil, s_proc=nil) { |entry| }  => self
  *
  * Perform a search, with the base DN +base_dn+, a scope of +scope+ and a
  * search filter of +filter+.
  *
  * If +attrs+ is present, it should be an array of the attributes that the
- * search should return. By default, all attributes are returned.
+ * search should return. By default, all attributes are returned, which is the
+ * same as specifying an empty array or *nil*. Alternatively, +attrs+ may be a
+ * single string, in which case it will be treated as a single element array.
  *
  * If +attrsonly+ is *true*, attributes will be returned, but not their values.
  *
@@ -985,13 +1067,18 @@ rb_ldap_conn_search2_b (VALUE rdata)
  * If +s_attr+ is given, it specifies the attribute on which to sort the
  * entries returned by the server. If +s_proc+ is given, it specifies a Proc
  * object that will be used to sort the entries returned by the server.
+ *
+ * Note that not all results may be returned by this method. If a
+ * size limit has been set for the number of results to be returned and this
+ * limit is exceeded, the results set will be truncated. You can check for
+ * this by calling LDAP::Conn#err immediately after this method and comparing
+ * the result to LDAP::LDAP_SIZELIMIT_EXCEEDED.
  */
 VALUE
 rb_ldap_conn_search2_s (int argc, VALUE argv[], VALUE self)
 {
   RB_LDAP_DATA *ldapdata;
   LDAPMessage *cmsg;
-  LDAPMessage *e;
   LDAP *cldap;
   VALUE ary = Qnil;
   VALUE rc_ary = Qnil;
@@ -1000,7 +1087,8 @@ rb_ldap_conn_search2_s (int argc, VALUE argv[], VALUE self)
   cldap = ldapdata->ldap;
 
   ary = rb_ary_new ();
-  if (ldapdata->err == LDAP_SUCCESS)
+  if (ldapdata->err == LDAP_SUCCESS
+      || ldapdata->err == LDAP_SIZELIMIT_EXCEEDED)
     {
       void *pass_data[] = { (void *) cldap, (void *) cmsg, (void *) ary };
 
@@ -1030,7 +1118,6 @@ rb_ldap_conn_search_ext_i (int argc, VALUE argv[], VALUE self,
   VALUE base, scope, filter, attrs, attrsonly;
   VALUE serverctrls, clientctrls, sec, usec, limit, s_attr, s_proc;
 
-  RB_LDAPENTRY_DATA *edata;
   LDAP *cldap;
   char *cbase;
   int cscope;
@@ -1039,8 +1126,7 @@ rb_ldap_conn_search_ext_i (int argc, VALUE argv[], VALUE self,
   char **cattrs;
   char *sort_attr;
   int cattrsonly;
-  VALUE msg;
-  int i, j;
+  int i;
   struct timeval tv;
   LDAPControl **sctrls, **cctrls;
 
@@ -1083,19 +1169,31 @@ rb_ldap_conn_search_ext_i (int argc, VALUE argv[], VALUE self,
     case 5:
       cattrsonly = (attrsonly == Qtrue) ? 1 : 0;
     case 4:
-      if (TYPE (attrs) == T_ARRAY)
+      if (TYPE (attrs) == T_NIL)
 	{
-	  cattrs = ALLOCA_N (char *, (RARRAY (attrs)->len + 1));
-	  for (i = 0; i < RARRAY (attrs)->len; i++)
-	    {
-	      cattrs[i] = StringValueCStr (RARRAY (attrs)->ptr[i]);
-	    };
-	  cattrs[RARRAY (attrs)->len] = NULL;
+	  cattrs = NULL;
 	}
       else
 	{
-	  cattrs = NULL;
-	};
+	  if (TYPE (attrs) == T_STRING)
+	    attrs = rb_ary_to_ary (attrs);
+	  else
+	    Check_Type (attrs, T_ARRAY);
+
+	  if (RARRAY (attrs)->len == 0)
+	    {
+	      cattrs = NULL;
+	    }
+	  else
+	    {
+	      cattrs = ALLOCA_N (char *, (RARRAY (attrs)->len + 1));
+	      for (i = 0; i < RARRAY (attrs)->len; i++)
+		{
+		  cattrs[i] = StringValueCStr (RARRAY (attrs)->ptr[i]);
+		};
+	      cattrs[RARRAY (attrs)->len] = NULL;
+	    }
+	}
     case 3:
       cbase = StringValueCStr (base);
       cscope = NUM2INT (scope);
@@ -1135,15 +1233,18 @@ rb_ldap_conn_search_ext_i (int argc, VALUE argv[], VALUE self,
 };
 
 /*
- * call-seq: conn.search_ext(base_dn, scope, filter, attrs=nil,
- *           attrsonly=false, sctrls, cctrls, sec=0, usec=0
- *           s_attr=nil, s_proc=nil) { |entry| }  => self
+ * call-seq:
+ * conn.search_ext(base_dn, scope, filter, attrs=nil, attrsonly=false,
+ *                 sctrls, cctrls, sec=0, usec=0, s_attr=nil, s_proc=nil)
+ *                 { |entry| }  => self
  *
  * Perform a search, with the base DN +base_dn+, a scope of +scope+ and a
  * search filter of +filter+.
  *
  * If +attrs+ is present, it should be an array of the attributes that the
- * search should return. By default, all attributes are returned.
+ * search should return. By default, all attributes are returned, which is the
+ * same as specifying an empty array or *nil*. Alternatively, +attrs+ may be a
+ * single string, in which case it will be treated as a single element array.
  *
  * If +attrsonly+ is *true*, attributes will be returned, but not their values.
  *
@@ -1156,6 +1257,12 @@ rb_ldap_conn_search_ext_i (int argc, VALUE argv[], VALUE self,
  * If +s_attr+ is given, it specifies the attribute on which to sort the
  * entries returned by the server. If +s_proc+ is given, it specifies a Proc
  * object that will be used to sort the entries returned by the server.
+ *
+ * Note that not all results may be returned by this method. If a
+ * size limit has been set for the number of results to be returned and this
+ * limit is exceeded, the results set will be truncated. You can check for
+ * this by calling LDAP::Conn#err immediately after this method and comparing
+ * the result to LDAP::LDAP_SIZELIMIT_EXCEEDED.
  */
 VALUE
 rb_ldap_conn_search_ext_s (int argc, VALUE argv[], VALUE self)
@@ -1163,12 +1270,12 @@ rb_ldap_conn_search_ext_s (int argc, VALUE argv[], VALUE self)
   RB_LDAP_DATA *ldapdata;
   LDAPMessage *cmsg;
   LDAP *cldap;
-  LDAPMessage *e;
 
   rb_ldap_conn_search_ext_i (argc, argv, self, &ldapdata, &cmsg);
   cldap = ldapdata->ldap;
 
-  if (ldapdata->err == LDAP_SUCCESS)
+  if (ldapdata->err == LDAP_SUCCESS
+      || ldapdata->err == LDAP_SIZELIMIT_EXCEEDED)
     {
       void *pass_data[] = { (void *) cldap, (void *) cmsg };
       rb_ensure (rb_ldap_conn_search_b, (VALUE) pass_data,
@@ -1181,14 +1288,19 @@ rb_ldap_conn_search_ext_s (int argc, VALUE argv[], VALUE self)
 /*
  * call-seq:
  * conn.search_ext2(base_dn, scope, filter, attrs=nil,
- * attrsonly=false, sctrls, cctrls, sec=0, usec=0, s_attr=nil,
- * s_proc=nil) { |entry| }  => self
+ *                  attrsonly=false, sctrls, cctrls, sec=0, usec=0,
+ *                  s_attr=nil, s_proc=nil)  => array
+ * conn.search_ext2(base_dn, scope, filter, attrs=nil,
+ *                  attrsonly=false, sctrls, cctrls, sec=0, usec=0,
+ *                  s_attr=nil, s_proc=nil) { |entry| }  => self
  *
  * Perform a search, with the base DN +base_dn+, a scope of +scope+ and a
  * search filter of +filter+.
  *
  * If +attrs+ is present, it should be an array of the attributes that the
- * search should return. By default, all attributes are returned.
+ * search should return. By default, all attributes are returned, which is the
+ * same as specifying an empty array or *nil*. Alternatively, +attrs+ may be a
+ * single string, in which case it will be treated as a single element array.
  *
  * If +attrsonly+ is *true*, attributes will be returned, but not their values.
  *
@@ -1201,13 +1313,18 @@ rb_ldap_conn_search_ext_s (int argc, VALUE argv[], VALUE self)
  * If +s_attr+ is given, it specifies the attribute on which to sort the
  * entries returned by the server. If +s_proc+ is given, it specifies a Proc
  * object that will be used to sort the entries returned by the server.
+ *
+ * Note that not all results may be returned by this method. If a
+ * size limit has been set for the number of results to be returned and this
+ * limit is exceeded, the results set will be truncated. You can check for
+ * this by calling LDAP::Conn#err immediately after this method and comparing
+ * the result to LDAP::LDAP_SIZELIMIT_EXCEEDED.
  */
 VALUE
 rb_ldap_conn_search_ext2_s (int argc, VALUE argv[], VALUE self)
 {
   RB_LDAP_DATA *ldapdata;
   LDAPMessage *cmsg;
-  LDAPMessage *e;
   LDAP *cldap;
   VALUE ary;
 
@@ -1215,7 +1332,8 @@ rb_ldap_conn_search_ext2_s (int argc, VALUE argv[], VALUE self)
   cldap = ldapdata->ldap;
 
   ary = rb_ary_new ();
-  if (ldapdata->err == LDAP_SUCCESS)
+  if (ldapdata->err == LDAP_SUCCESS
+      || ldapdata->err == LDAP_SIZELIMIT_EXCEEDED)
     {
       void *pass_data[] = { (void *) cldap, (void *) cmsg, (void *) ary };
       rb_ensure (rb_ldap_conn_search2_b, (VALUE) pass_data,
@@ -1234,7 +1352,8 @@ rb_ldap_conn_search_ext2_s (int argc, VALUE argv[], VALUE self)
 #endif
 
 /*
- * call-seq: conn.add(dn, attrs)  => self
+ * call-seq:
+ * conn.add(dn, attrs)  => self
  *
  * Add an entry with the DN, +dn+, and the attributes, +attrs+. +attrs+
  * should be either an array of LDAP#Mod objects or a hash of attribute/value
@@ -1283,7 +1402,8 @@ rb_ldap_conn_add_s (VALUE self, VALUE dn, VALUE attrs)
 
 #if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_ADD_EXT_S)
 /*
- * call-seq: conn.add_ext(dn, attrs, sctrls, cctrls)  => self
+ * call-seq:
+ * conn.add_ext(dn, attrs, sctrls, cctrls)  => self
  *
  * Add an entry with the DN, +dn+, and the attributes, +attrs+. +attrs+
  * should be either an array of LDAP#Mod objects or a hash of attribute/value
@@ -1338,7 +1458,8 @@ rb_ldap_conn_add_ext_s (VALUE self, VALUE dn, VALUE attrs,
 #endif
 
 /*
- * call-seq: conn.modify(dn, attrs)  => self
+ * call-seq:
+ * conn.modify(dn, attrs)  => self
  *
  * Modify an entry with the DN, +dn+, and the attributes, +attrs+. +attrs+
  * should be either an array of LDAP#Mod objects or a hash of attribute/value
@@ -1388,7 +1509,8 @@ rb_ldap_conn_modify_s (VALUE self, VALUE dn, VALUE attrs)
 
 #if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_MODIFY_EXT_S)
 /*
- * call-seq: conn.modify_ext(dn, attrs, sctrls, cctrls)  => self
+ * call-seq:
+ * conn.modify_ext(dn, attrs, sctrls, cctrls)  => self
  *
  * Modify an entry with the DN, +dn+, and the attributes, +attrs+. +attrs+
  * should be either an array of LDAP#Mod objects or a hash of attribute/value
@@ -1444,7 +1566,8 @@ rb_ldap_conn_modify_ext_s (VALUE self, VALUE dn, VALUE attrs,
 #endif
 
 /*
- * call-seq: conn.modrdn(dn, new_rdn, delete_old_rdn)  => self
+ * call-seq:
+ * conn.modrdn(dn, new_rdn, delete_old_rdn)  => self
  *
  * Modify the RDN of the entry with DN, +dn+, giving it the new RDN,
  * +new_rdn+. If +delete_old_rdn+ is *true*, the old RDN value will be deleted
@@ -1470,7 +1593,8 @@ rb_ldap_conn_modrdn_s (VALUE self, VALUE dn, VALUE newrdn, VALUE delete_p)
 };
 
 /*
- * call-seq: conn.delete(dn)  => self
+ * call-seq:
+ * conn.delete(dn)  => self
  *
  * Delete the entry with the DN, +dn+.
  */
@@ -1491,7 +1615,8 @@ rb_ldap_conn_delete_s (VALUE self, VALUE dn)
 
 #if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_DELETE_EXT_S)
 /*
- * call-seq: conn.delete_ext(dn, sctrls, cctrls)  => self
+ * call-seq:
+ * conn.delete_ext(dn, sctrls, cctrls)  => self
  *
  * Delete the entry with the DN, +dn+. +sctrls+ is an array of server
  * controls, whilst +cctrls+ is an array of client controls.
@@ -1518,7 +1643,8 @@ rb_ldap_conn_delete_ext_s (VALUE self, VALUE dn,
 
 #if defined(HAVE_LDAP_COMPARE_S)
 /*
- * call-seq: conn.compare(dn, attr, val)  => self
+ * call-seq:
+ * conn.compare(dn, attr, val)  => true or false
  *
  * Compare the DN given as +dn+ to see whether it has the attribute +attr+
  * with a value of +val+.
@@ -1535,15 +1661,24 @@ rb_ldap_conn_compare_s (VALUE self, VALUE dn, VALUE attr, VALUE val)
   c_val = StringValueCStr (val);
 
   ldapdata->err = ldap_compare_s (ldapdata->ldap, c_dn, c_attr, c_val);
+
+  if ((ldapdata->err) == LDAP_COMPARE_TRUE)
+    return Qtrue;
+  else if ((ldapdata->err) == LDAP_COMPARE_FALSE)
+    return Qfalse;
+
   Check_LDAP_Result (ldapdata->err);
 
+  fprintf (stderr, "rb_ldap_conn_compare_s() unexpectedly set no error.\n");
+
   return self;
 }
 #endif
 
 #if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_COMPARE_EXT_S)
 /*
- * call-seq: conn.compare_ext(dn, attr, val, sctrls, cctrls)  => self
+ * call-seq:
+ * conn.compare_ext(dn, attr, val, sctrls, cctrls)  => true or false
  *
  * Compare the DN given as +dn+ to see whether it has the attribute +attr+
  * with a value of +val+. +sctrls+ is an array of server controls, whilst
@@ -1555,25 +1690,49 @@ rb_ldap_conn_compare_ext_s (VALUE self, VALUE dn, VALUE attr, VALUE val,
 {
   RB_LDAP_DATA *ldapdata;
   char *c_dn, *c_attr;
+#ifdef USE_WLDAP32
+  char *c_val;
+#endif
   struct berval bval;
   LDAPControl **sctrls, **cctrls;
 
   GET_LDAP_DATA (self, ldapdata);
   c_dn = StringValueCStr (dn);
   c_attr = StringValueCStr (attr);
+#ifdef USE_WLDAP32
+  c_val = StringValueCStr (val);
+#endif
   bval.bv_val = StringValueCStr (val);
   bval.bv_len = RSTRING (val)->len;
   sctrls = rb_ldap_get_controls (serverctrls);
   cctrls = rb_ldap_get_controls (clientctrls);
 
-  ldapdata->err = ldap_compare_ext_s (ldapdata->ldap,
-				      c_dn, c_attr, &bval, sctrls, cctrls);
+  ldapdata->err = ldap_compare_ext_s (ldapdata->ldap, c_dn, c_attr,
+#ifdef USE_WLDAP32
+				      c_val,
+#endif
+				      &bval, sctrls, cctrls);
+
+  if ((ldapdata->err) == LDAP_COMPARE_TRUE)
+    return Qtrue;
+  else if ((ldapdata->err) == LDAP_COMPARE_FALSE)
+    return Qfalse;
+
   Check_LDAP_Result (ldapdata->err);
 
+  fprintf (stderr,
+	   "rb_ldap_conn_compare_ext_s() unexpectedly set no error.\n");
+
   return self;
 }
 #endif
 
+/*
+ * call-seq:
+ * conn.err  => Fixnum
+ *
+ * Return the error associated with the most recent LDAP operation.
+ */
 VALUE
 rb_ldap_conn_err (VALUE self)
 {
