@@ -1,7 +1,9 @@
-/* -*- C -*-
- * $Id: winldap.h,v 1.3 2002/11/18 08:22:31 ttate Exp $
+/*
+ * $Id: winldap.h,v 1.5 2006/04/20 16:34:21 ianmacd Exp $
+ *
  * Copyright (C) 2001 Takaaki Tateishi <ttate@kt.jaist.ac.jp>
- * References: MSDN Library, OpenLDAP, Cygwin
+ * Copyright (C) 2006 Ian Macdonald <ian@caliban.org>
+ *
  */
 
 #ifndef WINLDAP_H
@@ -183,9 +185,7 @@ DS_IS_DNS_NAME
 */
 
 
-struct ldap {
-  /* We don't know how the structure is defined. */
-};
+struct ldap;
 typedef struct ldap LDAP, *PLDAP;
 
 typedef struct ldapcontrol {
@@ -194,9 +194,7 @@ typedef struct ldapcontrol {
   BOOLEAN ldctl_iscritical;
 } LDAPControl, *PLDAPControl;
 
-struct ldapmsg {
-  /* We don't know how the structure is defined. */
-};
+struct ldapmsg;
 typedef struct ldapmsg LDAPMessage, *PLDAPMessage;
 
 
@@ -225,18 +223,11 @@ typedef struct LdapReferralCallback {
 } LDAP_REFERRAL_CALLBACK, *PLDAP_REFERRAL_CALLBACK;
 */
 
-struct ldapsearch {
-  /* We don't know how the structure is defined. */
-};
+struct ldapsearch;
 typedef struct ldapsearch LDAPSearch, *PLDAPSearch;
 
 #if defined(HAVE_SYS_TIME_H)
 # include <sys/time.h>
-#else
-struct timeval {
-  long tv_sec;
-  long tv_usec;
-};
 #endif
 typedef struct timeval LDAP_TIMEVAL;
 typedef struct timeval *PLDAP_TIMEVAL;
@@ -281,7 +272,8 @@ ULONG ldap_add_ext_s(LDAP *ld, PCHAR dn, LDAPMod *attrs[],
 		     LDAPControl **sctls, LDAPControl **cctls);
 ULONG ldap_compare_s(LDAP *ld, PCHAR dn, PCHAR attr, PCHAR value);
 ULONG ldap_compare_ext_s(LDAP *ld, PCHAR dn, PCHAR attr, PCHAR value,
-			 LDAPControl **sctls, LDAPControl **cctls);
+			 struct berval *data, LDAPControl **sctls,
+			 LDAPControl **cctls);
 ULONG ldap_count_entries(LDAP *ld, LDAPMessage *res);
 ULONG ldap_count_values(PCHAR *vals);
 ULONG ldap_delete_s(LDAP *ld,  PCHAR dn);
